package java.util.concurrent;

import java.util.concurrent.locks.LockSupport;
import sun.misc.Unsafe;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Spliterators;
import java.util.Spliterator;
import java.util.Collections;
import java.util.Iterator;
import java.util.Collection;
import java.util.concurrent.locks.ReentrantLock;
import java.io.Serializable;
import java.util.AbstractQueue;

public class SynchronousQueue<E> extends AbstractQueue<E> implements BlockingQueue<E>, Serializable
{
    private static final long serialVersionUID = -3223113410248163686L;
    static final int NCPUS;
    static final int maxTimedSpins;
    static final int maxUntimedSpins;
    static final long spinForTimeoutThreshold = 1000L;
    private transient volatile SynchronousQueue.Transferer<E> transferer;
    private ReentrantLock qlock;
    private SynchronousQueue.WaitQueue waitingProducers;
    private SynchronousQueue.WaitQueue waitingConsumers;
    
    public SynchronousQueue() {
        this(false);
    }
    
    public SynchronousQueue(boolean b) {
        this.transferer = (b ? new SynchronousQueue.TransferQueue() : new SynchronousQueue.TransferStack());
    }
    
    @Override
    public void put(E e) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        if (this.transferer.transfer(e, false, 0L) == null) {
            Thread.interrupted();
            throw new InterruptedException();
        }
        return;
    }
    
    @Override
    public boolean offer(E e, long n, TimeUnit timeUnit) throws InterruptedException {
        if (e == null) {
            throw new NullPointerException();
        }
        if (this.transferer.transfer(e, true, timeUnit.toNanos(n)) != null) {
            return true;
        }
        if (!Thread.interrupted()) {
            return false;
        }
        throw new InterruptedException();
    }
    
    @Override
    public boolean offer(E e) {
        if (e == null) {
            throw new NullPointerException();
        }
        return this.transferer.transfer(e, true, 0L) != null;
    }
    
    @Override
    public E take() throws InterruptedException {
        Object transfer = this.transferer.transfer(null, false, 0L);
        if (transfer != null) {
            return transfer;
        }
        Thread.interrupted();
        throw new InterruptedException();
    }
    
    @Override
    public E poll(long n, TimeUnit timeUnit) throws InterruptedException {
        Object transfer = this.transferer.transfer(null, true, timeUnit.toNanos(n));
        if (transfer != null || !Thread.interrupted()) {
            return transfer;
        }
        throw new InterruptedException();
    }
    
    @Override
    public E poll() {
        return this.transferer.transfer(null, true, 0L);
    }
    
    @Override
    public boolean isEmpty() {
        return true;
    }
    
    @Override
    public int size() {
        return 0;
    }
    
    @Override
    public int remainingCapacity() {
        return 0;
    }
    
    @Override
    public void clear() {
    }
    
    @Override
    public boolean contains(Object o) {
        return false;
    }
    
    @Override
    public boolean remove(Object o) {
        return false;
    }
    
    @Override
    public boolean containsAll(Collection<?> collection) {
        return collection.isEmpty();
    }
    
    @Override
    public boolean removeAll(Collection<?> collection) {
        return false;
    }
    
    @Override
    public boolean retainAll(Collection<?> collection) {
        return false;
    }
    
    @Override
    public E peek() {
        return null;
    }
    
    @Override
    public Iterator<E> iterator() {
        return Collections.emptyIterator();
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return Spliterators.emptySpliterator();
    }
    
    @Override
    public Object[] toArray() {
        return new Object[0];
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        if (array.length > 0) {
            array[0] = null;
            return array;
        }
        return array;
    }
    
    @Override
    public int drainTo(Collection<? super E> collection) {
        if (collection == null) {
            throw new NullPointerException();
        }
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        int n;
        Object poll;
        for (n = 0; (poll = this.poll()) != null; ++n) {
            collection.add(poll);
        }
        return n;
    }
    
    @Override
    public int drainTo(Collection<? super E> collection, int n) {
        if (collection == null) {
            throw new NullPointerException();
        }
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        int n2;
        Object poll;
        for (n2 = 0; n2 < n && (poll = this.poll()) != null; ++n2) {
            collection.add(poll);
        }
        return n2;
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        if ((this.transferer instanceof SynchronousQueue.TransferQueue)) {
            this.qlock = new ReentrantLock(true);
            this.waitingProducers = new SynchronousQueue.FifoWaitQueue();
            this.waitingConsumers = new SynchronousQueue.FifoWaitQueue();
        }
        else {
            this.qlock = new ReentrantLock();
            this.waitingProducers = new SynchronousQueue.LifoWaitQueue();
            this.waitingConsumers = new SynchronousQueue.LifoWaitQueue();
        }
        objectOutputStream.defaultWriteObject();
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        if ((this.waitingProducers instanceof SynchronousQueue.FifoWaitQueue)) {
            this.transferer = new SynchronousQueue.TransferQueue();
        }
        else {
            this.transferer = new SynchronousQueue.TransferStack();
        }
    }
    
    static long objectFieldOffset(Unsafe unsafe, String s, Class<?> clazz) {
        try {
            return unsafe.objectFieldOffset(clazz.getDeclaredField(s));
        }
        catch (NoSuchFieldException cause) {
            NoSuchFieldError noSuchFieldError = new NoSuchFieldError(s);
            noSuchFieldError.initCause(cause);
            throw noSuchFieldError;
        }
    }
    
    static {
        NCPUS = Runtime.getRuntime().availableProcessors();
        maxTimedSpins = ((SynchronousQueue.NCPUS < 2) ? 0 : 32);
        maxUntimedSpins = SynchronousQueue.maxTimedSpins * 16;
    }
    
    abstract static class Transferer<E>
    {
        abstract E transfer(E p0, boolean p1, long p2);
    }
    
    static final class TransferStack<E> extends SynchronousQueue.Transferer<E>
    {
        static final int REQUEST = 0;
        static final int DATA = 1;
        static final int FULFILLING = 2;
        volatile SynchronousQueue.TransferStack.SNode head;
        private static final Unsafe UNSAFE;
        private static final long headOffset;
        
        static boolean isFulfilling(int n) {
            return (n & 0x2) != 0x0;
        }
        
        boolean casHead(SynchronousQueue.TransferStack.SNode expected, SynchronousQueue.TransferStack.SNode x) {
            return expected == this.head && SynchronousQueue.TransferStack.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferStack.headOffset, expected, x);
        }
        
        static SynchronousQueue.TransferStack.SNode snode(SynchronousQueue.TransferStack.SNode sNode, Object o, SynchronousQueue.TransferStack.SNode next, int mode) {
            if (sNode == null) {
                sNode = new SNode(o);
            }
            sNode.mode = mode;
            sNode.next = next;
            return sNode;
        }
        
        @Override
        E transfer(E e, boolean b, long n) {
            SynchronousQueue.TransferStack.SNode sNode = null;
            int n2 = e != null;
            while (true) {
                SynchronousQueue.TransferStack.SNode head = this.head;
                if (head == null || head.mode == n2) {
                    if (b && n <= 0L) {
                        if (head != null && head.isCancelled()) {
                            this.casHead(head, head.next);
                        }
                        else {
                            return null;
                        }
                    }
                    else if (this.casHead(head, sNode = snode(sNode, e, head, n2))) {
                        SynchronousQueue.TransferStack.SNode awaitFulfill = this.awaitFulfill(sNode, b, n);
                        if (awaitFulfill == sNode) {
                            this.clean(sNode);
                            return null;
                        }
                        SynchronousQueue.TransferStack.SNode head2;
                        if ((head2 = this.head) != null && head2.next == sNode) {
                            this.casHead(head2, sNode.next);
                        }
                        return (n2 == 0) ? awaitFulfill.item : sNode.item;
                    }
                    else {
                        continue;
                    }
                }
                else if (!isFulfilling(head.mode)) {
                    if (head.isCancelled()) {
                        this.casHead(head, head.next);
                    }
                    else if (this.casHead(head, sNode = snode(sNode, e, head, 0x2 | n2))) {
                        while (true) {
                            SynchronousQueue.TransferStack.SNode next3 = sNode.next;
                            if (next3 == null) {
                                this.casHead(sNode, null);
                                sNode = null;
                                break;
                            }
                            SynchronousQueue.TransferStack.SNode next4 = next3.next;
                            if (next3.tryMatch(sNode)) {
                                this.casHead(sNode, next4);
                                return (n2 == 0) ? next3.item : sNode.item;
                            }
                            sNode.casNext(next3, next4);
                        }
                    }
                    else {
                        continue;
                    }
                }
                else {
                    SynchronousQueue.TransferStack.SNode next = head.next;
                    if (next == null) {
                        this.casHead(head, null);
                    }
                    else {
                        SynchronousQueue.TransferStack.SNode next2 = next.next;
                        if (next.tryMatch(head)) {
                            this.casHead(head, next2);
                        }
                        else {
                            head.casNext(next, next2);
                        }
                    }
                }
            }
        }
        
        SynchronousQueue.TransferStack.SNode awaitFulfill(SynchronousQueue.TransferStack.SNode sNode, boolean b, long nanos) {
            long n = b ? (System.nanoTime() + nanos) : 0L;
            Thread currentThread = Thread.currentThread();
            int n2 = this.shouldSpin(sNode) ? (b ? SynchronousQueue.maxTimedSpins : SynchronousQueue.maxUntimedSpins) : 0;
            SynchronousQueue.TransferStack.SNode match;
            while (true) {
                if (currentThread.isInterrupted()) {
                    sNode.tryCancel();
                }
                match = sNode.match;
                if (match != null) {
                    break;
                }
                if (b) {
                    nanos = n - System.nanoTime();
                    if (nanos <= 0L) {
                        sNode.tryCancel();
                        continue;
                    }
                }
                if (n2 > 0) {
                    n2 = (this.shouldSpin(sNode) ? (n2 - 1) : 0);
                }
                else if (sNode.waiter == null) {
                    sNode.waiter = currentThread;
                }
                else if (!b) {
                    LockSupport.park(this);
                }
                else if (nanos > 1000L) {
                    LockSupport.parkNanos(this, nanos);
                }
                else {
                    continue;
                }
            }
            return match;
        }
        
        boolean shouldSpin(SynchronousQueue.TransferStack.SNode sNode) {
            SynchronousQueue.TransferStack.SNode head = this.head;
            return head == sNode || head == null || isFulfilling(head.mode);
        }
        
        void clean(SynchronousQueue.TransferStack.SNode sNode) {
            sNode.item = null;
            sNode.waiter = null;
            SynchronousQueue.TransferStack.SNode sNode2 = sNode.next;
            if (sNode2 != null && sNode2.isCancelled()) {
                sNode2 = sNode2.next;
            }
            SynchronousQueue.TransferStack.SNode head;
            while ((head = this.head) != null && head != sNode2 && head.isCancelled()) {
                this.casHead(head, head.next);
            }
            while (head != null && head != sNode2) {
                SynchronousQueue.TransferStack.SNode next = head.next;
                if (next != null && next.isCancelled()) {
                    head.casNext(next, next.next);
                }
                else {
                    head = next;
                }
            }
        }
        
        static {
            try {
                UNSAFE = Unsafe.getUnsafe();
                headOffset = SynchronousQueue.TransferStack.UNSAFE.objectFieldOffset(SynchronousQueue.TransferStack.class.getDeclaredField("head"));
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
        }
        
        static final class SNode
        {
            volatile SynchronousQueue.TransferStack.SNode next;
            volatile SynchronousQueue.TransferStack.SNode match;
            volatile Thread waiter;
            Object item;
            int mode;
            private static final Unsafe UNSAFE;
            private static final long matchOffset;
            private static final long nextOffset;
            
            SNode(Object item) {
                this.item = item;
            }
            
            boolean casNext(SynchronousQueue.TransferStack.SNode expected, SynchronousQueue.TransferStack.SNode x) {
                return expected == this.next && SynchronousQueue.TransferStack.SNode.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferStack.SNode.nextOffset, expected, x);
            }
            
            boolean tryMatch(SynchronousQueue.TransferStack.SNode x) {
                if (this.match == null && SynchronousQueue.TransferStack.SNode.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferStack.SNode.matchOffset, null, x)) {
                    Thread waiter = this.waiter;
                    if (waiter != null) {
                        this.waiter = null;
                        LockSupport.unpark(waiter);
                    }
                    return true;
                }
                else {
                    return this.match == x;
                }
            }
            
            void tryCancel() {
                SynchronousQueue.TransferStack.SNode.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferStack.SNode.matchOffset, null, this);
            }
            
            boolean isCancelled() {
                return this.match == this;
            }
            
            static {
                try {
                    UNSAFE = Unsafe.getUnsafe();
                    Class<Object> clazz = SynchronousQueue.TransferStack.SNode.class;
                    matchOffset = SynchronousQueue.TransferStack.SNode.UNSAFE.objectFieldOffset(clazz.getDeclaredField("match"));
                    nextOffset = SynchronousQueue.TransferStack.SNode.UNSAFE.objectFieldOffset(clazz.getDeclaredField("next"));
                }
                catch (Exception cause) {
                    throw new Error(cause);
                }
            }
        }
    }
    
    static final class TransferQueue<E> extends SynchronousQueue.Transferer<E>
    {
        transient volatile SynchronousQueue.TransferQueue.QNode head;
        transient volatile SynchronousQueue.TransferQueue.QNode tail;
        transient volatile SynchronousQueue.TransferQueue.QNode cleanMe;
        private static final Unsafe UNSAFE;
        private static final long headOffset;
        private static final long tailOffset;
        private static final long cleanMeOffset;
        
        TransferQueue() {
            SynchronousQueue.TransferQueue.QNode qNode = new SynchronousQueue.TransferQueue.QNode(null, false);
            this.head = qNode;
            this.tail = qNode;
        }
        
        void advanceHead(SynchronousQueue.TransferQueue.QNode qNode, SynchronousQueue.TransferQueue.QNode x) {
            if (qNode == this.head && SynchronousQueue.TransferQueue.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferQueue.headOffset, qNode, x)) {
                qNode.next = qNode;
            }
        }
        
        void advanceTail(SynchronousQueue.TransferQueue.QNode expected, SynchronousQueue.TransferQueue.QNode x) {
            if (this.tail == expected) {
                SynchronousQueue.TransferQueue.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferQueue.tailOffset, expected, x);
            }
        }
        
        boolean casCleanMe(SynchronousQueue.TransferQueue.QNode expected, SynchronousQueue.TransferQueue.QNode x) {
            return this.cleanMe == expected && SynchronousQueue.TransferQueue.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferQueue.cleanMeOffset, expected, x);
        }
        
        @Override
        E transfer(E e, boolean b, long n) {
            Object item = null;
            int n2 = e != null;
            while (true) {
                SynchronousQueue.TransferQueue.QNode tail = this.tail;
                SynchronousQueue.TransferQueue.QNode head = this.head;
                if (tail == null) {
                    continue;
                }
                if (head == null) {
                    continue;
                }
                if (head == tail || (tail.isData ? 1 : 0) == n2) {
                    SynchronousQueue.TransferQueue.QNode next2 = tail.next;
                    if (tail != this.tail) {
                        continue;
                    }
                    if (next2 != null) {
                        this.advanceTail(tail, next2);
                    }
                    else {
                        if (b && n <= 0L) {
                            return null;
                        }
                        if (item == null) {
                            item = new SynchronousQueue.TransferQueue.QNode(e, (boolean)n2);
                        }
                        if (!tail.casNext(null, item)) {
                            continue;
                        }
                        this.advanceTail(tail, item);
                        Object awaitFulfill = this.awaitFulfill(item, e, b, n);
                        if (awaitFulfill == item) {
                            this.clean(tail, item);
                            return null;
                        }
                        if (!item.isOffList()) {
                            this.advanceHead(tail, item);
                            if (awaitFulfill != null) {
                                item.item = item;
                            }
                            item.waiter = null;
                        }
                        return (awaitFulfill != null) ? awaitFulfill : e;
                    }
                }
                else {
                    SynchronousQueue.TransferQueue.QNode next = head.next;
                    if (tail != this.tail || next == null) {
                        continue;
                    }
                    if (head != this.head) {
                        continue;
                    }
                    Object item2 = next.item;
                    if (n2 == ((item2 != null) ? 1 : 0) || item2 == next || !next.casItem(item2, e)) {
                        this.advanceHead(head, next);
                    }
                    else {
                        this.advanceHead(head, next);
                        LockSupport.unpark(next.waiter);
                        return (item2 != null) ? item2 : e;
                    }
                }
            }
        }
        
        Object awaitFulfill(SynchronousQueue.TransferQueue.QNode qNode, E e, boolean b, long nanos) {
            long n = b ? (System.nanoTime() + nanos) : 0L;
            Thread currentThread = Thread.currentThread();
            int n2 = (this.head.next == qNode) ? (b ? SynchronousQueue.maxTimedSpins : SynchronousQueue.maxUntimedSpins) : 0;
            Object item;
            while (true) {
                if (currentThread.isInterrupted()) {
                    qNode.tryCancel(e);
                }
                item = qNode.item;
                if (item != e) {
                    break;
                }
                if (b) {
                    nanos = n - System.nanoTime();
                    if (nanos <= 0L) {
                        qNode.tryCancel(e);
                        continue;
                    }
                }
                if (n2 > 0) {
                    --n2;
                }
                else if (qNode.waiter == null) {
                    qNode.waiter = currentThread;
                }
                else if (!b) {
                    LockSupport.park(this);
                }
                else if (nanos > 1000L) {
                    LockSupport.parkNanos(this, nanos);
                }
                else {
                    continue;
                }
            }
            return item;
        }
        
        void clean(SynchronousQueue.TransferQueue.QNode qNode, SynchronousQueue.TransferQueue.QNode qNode2) {
            qNode2.waiter = null;
            while (qNode.next == qNode2) {
                SynchronousQueue.TransferQueue.QNode head = this.head;
                SynchronousQueue.TransferQueue.QNode next = head.next;
                if (next != null && next.isCancelled()) {
                    this.advanceHead(head, next);
                }
                else {
                    SynchronousQueue.TransferQueue.QNode tail = this.tail;
                    if (tail == head) {
                        return;
                    }
                    SynchronousQueue.TransferQueue.QNode next2 = tail.next;
                    if (tail != this.tail) {
                        continue;
                    }
                    if (next2 != null) {
                        this.advanceTail(tail, next2);
                    }
                    else {
                        if (qNode2 != tail) {
                            SynchronousQueue.TransferQueue.QNode next3 = qNode2.next;
                            if (next3 == qNode2 || qNode.casNext(qNode2, next3)) {
                                return;
                            }
                        }
                        SynchronousQueue.TransferQueue.QNode cleanMe = this.cleanMe;
                        if (cleanMe != null) {
                            SynchronousQueue.TransferQueue.QNode next4 = cleanMe.next;
                            SynchronousQueue.TransferQueue.QNode next5;
                            if (next4 == null || next4 == cleanMe || !next4.isCancelled() || (next4 != tail && (next5 = next4.next) != null && next5 != next4 && cleanMe.casNext(next4, next5))) {
                                this.casCleanMe(cleanMe, null);
                            }
                            if (cleanMe == qNode) {
                                return;
                            }
                            continue;
                        }
                        else {
                            if (this.casCleanMe(null, qNode)) {
                                return;
                            }
                            continue;
                        }
                    }
                }
            }
        }
        
        static {
            try {
                UNSAFE = Unsafe.getUnsafe();
                Class<Object> clazz = SynchronousQueue.TransferQueue.class;
                headOffset = SynchronousQueue.TransferQueue.UNSAFE.objectFieldOffset(clazz.getDeclaredField("head"));
                tailOffset = SynchronousQueue.TransferQueue.UNSAFE.objectFieldOffset(clazz.getDeclaredField("tail"));
                cleanMeOffset = SynchronousQueue.TransferQueue.UNSAFE.objectFieldOffset(clazz.getDeclaredField("cleanMe"));
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
        }
        
        static final class QNode
        {
            volatile SynchronousQueue.TransferQueue.QNode next;
            volatile Object item;
            volatile Thread waiter;
            final boolean isData;
            private static final Unsafe UNSAFE;
            private static final long itemOffset;
            private static final long nextOffset;
            
            QNode(Object item, boolean isData) {
                this.item = item;
                this.isData = isData;
            }
            
            boolean casNext(SynchronousQueue.TransferQueue.QNode expected, SynchronousQueue.TransferQueue.QNode x) {
                return this.next == expected && SynchronousQueue.TransferQueue.QNode.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferQueue.QNode.nextOffset, expected, x);
            }
            
            boolean casItem(Object expected, Object x) {
                return this.item == expected && SynchronousQueue.TransferQueue.QNode.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferQueue.QNode.itemOffset, expected, x);
            }
            
            void tryCancel(Object expected) {
                SynchronousQueue.TransferQueue.QNode.UNSAFE.compareAndSwapObject(this, SynchronousQueue.TransferQueue.QNode.itemOffset, expected, this);
            }
            
            boolean isCancelled() {
                return this.item == this;
            }
            
            boolean isOffList() {
                return this.next == this;
            }
            
            static {
                try {
                    UNSAFE = Unsafe.getUnsafe();
                    Class<Object> clazz = SynchronousQueue.TransferQueue.QNode.class;
                    itemOffset = SynchronousQueue.TransferQueue.QNode.UNSAFE.objectFieldOffset(clazz.getDeclaredField("item"));
                    nextOffset = SynchronousQueue.TransferQueue.QNode.UNSAFE.objectFieldOffset(clazz.getDeclaredField("next"));
                }
                catch (Exception cause) {
                    throw new Error(cause);
                }
            }
        }
    }
    
    static class WaitQueue implements Serializable
    {
    }
    
    static class LifoWaitQueue extends SynchronousQueue.WaitQueue
    {
        private static final long serialVersionUID = -3633113410248163686L;
    }
    
    static class FifoWaitQueue extends SynchronousQueue.WaitQueue
    {
        private static final long serialVersionUID = -3623113410248163686L;
    }
}
